/*
Copyright © 2023 Philipp Wolfer <phw@uploadedlobster.com>

This file is part of Scotty.

Scotty is free software: you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later version.

Scotty is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
Scotty. If not, see <https://www.gnu.org/licenses/>.
*/
package listenbrainz_test

import (
	"testing"
	"time"

	"github.com/spf13/viper"
	"github.com/stretchr/testify/assert"
	"github.com/stretchr/testify/require"
	"go.uploadedlobster.com/scotty/backends/listenbrainz"
	"go.uploadedlobster.com/scotty/models"
)

func TestFromConfig(t *testing.T) {
	config := viper.New()
	config.Set("token", "thetoken")
	backend := (&listenbrainz.ListenBrainzApiBackend{}).FromConfig(config)
	assert.IsType(t, &listenbrainz.ListenBrainzApiBackend{}, backend)
}

func TestListenBrainzListenAsListen(t *testing.T) {
	lbListen := listenbrainz.Listen{
		ListenedAt: 1699289873,
		UserName:   "outsidecontext",
		TrackMetadata: listenbrainz.Track{
			TrackName:   "The Track",
			ArtistName:  "Dool",
			ReleaseName: "Here Now, There Then",
			AdditionalInfo: map[string]any{
				"duration_ms":        413787,
				"foo":                "bar",
				"isrc":               "DES561620801",
				"tracknumber":        5,
				"discnumber":         1,
				"recording_mbid":     "c0a1fc94-5f04-4a5f-bc09-e5de0c49cd12",
				"release_group_mbid": "80aca1ee-aa51-41be-9f75-024710d92ff4",
				"release_mbid":       "d7f22677-9803-4d21-ba42-081b633a6f68",
			},
		},
	}
	listen := lbListen.AsListen()
	assert.Equal(t, time.Unix(1699289873, 0), listen.ListenedAt)
	assert.Equal(t, lbListen.UserName, listen.UserName)
	assert.Equal(t, time.Duration(413787*time.Millisecond), listen.Duration)
	assert.Equal(t, lbListen.TrackMetadata.TrackName, listen.TrackName)
	assert.Equal(t, lbListen.TrackMetadata.ReleaseName, listen.ReleaseName)
	assert.Equal(t, []string{lbListen.TrackMetadata.ArtistName}, listen.ArtistNames)
	assert.Equal(t, 5, listen.TrackNumber)
	assert.Equal(t, 1, listen.DiscNumber)
	assert.Equal(t, models.MBID("c0a1fc94-5f04-4a5f-bc09-e5de0c49cd12"), listen.RecordingMbid)
	assert.Equal(t, models.MBID("d7f22677-9803-4d21-ba42-081b633a6f68"), listen.ReleaseMbid)
	assert.Equal(t, models.MBID("80aca1ee-aa51-41be-9f75-024710d92ff4"), listen.ReleaseGroupMbid)
	assert.Equal(t, "DES561620801", listen.ISRC)
	assert.Equal(t, lbListen.TrackMetadata.AdditionalInfo["foo"], listen.AdditionalInfo["foo"])
}

func TestListenBrainzFeedbackAsLove(t *testing.T) {
	recordingMbid := "c0a1fc94-5f04-4a5f-bc09-e5de0c49cd12"
	releaseMbid := "d7f22677-9803-4d21-ba42-081b633a6f68"
	artistMbid := "d7f22677-9803-4d21-ba42-081b633a6f68"
	feedback := listenbrainz.Feedback{
		Created:       1699859066,
		RecordingMbid: recordingMbid,
		Score:         1,
		UserName:      "ousidecontext",
		TrackMetadata: &listenbrainz.Track{
			TrackName:   "Oweynagat",
			ArtistName:  "Dool",
			ReleaseName: "Here Now, There Then",
			MbidMapping: &listenbrainz.MbidMapping{
				RecordingMbid: recordingMbid,
				ReleaseMbid:   releaseMbid,
				ArtistMbids:   []string{artistMbid},
			},
		},
	}
	love := feedback.AsLove()
	assert := assert.New(t)
	assert.Equal(time.Unix(1699859066, 0).Unix(), love.Created.Unix())
	assert.Equal(feedback.UserName, love.UserName)
	assert.Equal(feedback.TrackMetadata.TrackName, love.TrackName)
	assert.Equal(feedback.TrackMetadata.ReleaseName, love.ReleaseName)
	assert.Equal([]string{feedback.TrackMetadata.ArtistName}, love.ArtistNames)
	assert.Equal(models.MBID(recordingMbid), love.RecordingMbid)
	assert.Equal(models.MBID(recordingMbid), love.Track.RecordingMbid)
	assert.Equal(models.MBID(releaseMbid), love.Track.ReleaseMbid)
	require.Len(t, love.Track.ArtistMbids, 1)
	assert.Equal(models.MBID(artistMbid), love.Track.ArtistMbids[0])
}

func TestListenBrainzPartialFeedbackAsLove(t *testing.T) {
	recordingMbid := "c0a1fc94-5f04-4a5f-bc09-e5de0c49cd12"
	feedback := listenbrainz.Feedback{
		Created:       1699859066,
		RecordingMbid: recordingMbid,
		Score:         1,
	}
	love := feedback.AsLove()
	assert := assert.New(t)
	assert.Equal(time.Unix(1699859066, 0).Unix(), love.Created.Unix())
	assert.Equal(models.MBID(recordingMbid), love.RecordingMbid)
	assert.Equal(models.MBID(recordingMbid), love.Track.RecordingMbid)
	assert.Empty(love.Track.TrackName)
}
